package edu.unca.cs.csci202;

import java.util.* ;

/**
 *
 * @author J Dean Brock <brock@unca.edu>
 */
public class SimpleEvalExpr {

    // A simple tokenizer
    // Parse an ordinary expression, such as "(13+17)*19+21"
    // into infix list such as [ ( 13 + 17 ) * 19 + 21 ]
    private static Deque<Object> string2infix(String expr)
             throws IllegalArgumentException {
        // The "diamond" operator on LinkedList is new to Java 7.
        // You may need to remove it for Java 6.
        Deque<Object> rpnExp = new LinkedList<>() ;
        int i = 0 ;
        // The push calls all use autoboxing
        while (i<expr.length()) {
            while (i<expr.length() && Character.isWhitespace(expr.charAt(i))) {
                ++i ;
            }
            if (i<expr.length()) {
                char nextC = expr.charAt(i) ;
                if (nextC == '(') {
                    // This one doesn't use autoboxing, just like if Java 1.5
                    rpnExp.push((Character)nextC) ;
                } else if (nextC == ')') {
                    rpnExp.push(nextC) ;
                } else if (nextC == '+') {
                    rpnExp.push(nextC) ;
                } else if (nextC == '*') {
                    rpnExp.push(nextC) ;
                } else if (Character.isDigit(nextC)) {
                    int pNum = Character.digit(nextC, 10) ;
                    while (i+1<expr.length() &&
                            Character.isDigit(expr.charAt(i+1))) {
                        ++i ;
                        pNum = 10*pNum + Character.digit(expr.charAt(i), 10) ;
                    }
                    rpnExp.push(pNum) ;
                } else {
                    throw new IllegalArgumentException("Bad character") ;
                }
            }
            ++i ;
        }
        return rpnExp ;
    }

    // This is a complicated routine that receives an
    // expression in the infix form and checks if it
    // is valid. Instead of returning a boolean, it signals
    // IllegalArgumentException of invalid expressions
    private static void syntaxCheck (Deque<Object> infix)
            throws IllegalArgumentException {
        // The opCountStack and valCountStack are used to keep
        // up with some interesting things between paraentheses
        Deque<Integer> opCountStack  = new LinkedList<>() ;
        Deque<Integer> valCountStack = new LinkedList<>() ;
        opCountStack.push(0);
        valCountStack.push(0);
        // Iterators are so useful
        Iterator<Object> revStack = infix.descendingIterator() ;
        while (revStack.hasNext()) {
            Object token = revStack.next() ;
            // used so much, let's get their values now
            int presOpCount  = opCountStack.peek() ;
            int presValCount = valCountStack.peek() ;
            if (token instanceof Character) {
                // auto unboxing occurs here
                char nextC = (Character)token ;
                if (nextC == '(') {
                    // legit only if numbers of values and operations are equal
                    if (presValCount != presOpCount) {
                        throw new IllegalArgumentException("Problem at (") ;
                    }
                    // Push new ( level
                    opCountStack.push(0) ;
                    valCountStack.push(0) ;
                } else if (nextC == ')') {
                    // legit only if one more value than operation
                    if (presValCount != presOpCount+1) {
                        throw new IllegalArgumentException("Problem at )") ;
                    }
                    // pop the stacks
                    opCountStack.pop() ;
                    valCountStack.pop() ;
                    // Increment value count
                    try {
                        valCountStack.push(valCountStack.pop()+1) ;
                    // Watch out for too many )'s
                    } catch (NoSuchElementException nsee) {
                        System.out.println(valCountStack) ;
                        throw new IllegalArgumentException("Unpaired )") ;
                    }
                } else /* nextC is '+' or '*' */ {
                    // legit only if one more value than operation
                    if (presValCount != presOpCount + 1) {
                        throw new IllegalArgumentException("Problem at " +
                                nextC) ;
                    }
                    opCountStack.push(opCountStack.pop()+1) ;
                }
            } else /* must be a value! */ {
                // legit only if numbers of values and operations are equal
                if (presValCount != presOpCount) {
                    throw new IllegalArgumentException("Problem at " +
                            (int)token) ;
                }
                valCountStack.push(valCountStack.pop()+1) ;
            }
        }
        if (valCountStack.size() != 1) {
            throw new IllegalArgumentException("Unpaired (") ;
        }
        if (valCountStack.peek() != opCountStack.peek()+1) {
            throw new IllegalArgumentException("Bad format") ;
        }
    }
    
    // Transform an infix list such as [ ( 13 + 17 ) * 19 + 21 ]
    // into a postfix list, such as [ 13 17 + 19 * 21 + ]
    // The algorithm on page 375 to 378 has a similar function
    private static Deque<Object> infix2postfix(Deque<Object> infix) {
        Deque<Object> postfix = new LinkedList<>() ;
        // Pending operator stack
        Deque<Character> pendingOps = new LinkedList<>() ;
        pendingOps.push('(') ;
        Iterator<Object> revStack = infix.descendingIterator() ;
        while (revStack.hasNext()) {
            Object token = revStack.next() ;
            if (token instanceof Character) {
                char nextC = (Character)token ;
                if (nextC == ')') {
                    while(pendingOps.peek() != '(') {
                        postfix.push(pendingOps.pop()) ;
                    }
                    pendingOps.pop() ;
                } else if (nextC == '(') {
                    pendingOps.push('(') ;
                } else if (nextC == '*') {
                    pendingOps.push('*') ;
                } else /* nextC == '+' */ {
                    while(pendingOps.peek() == '*') {
                        postfix.push(pendingOps.pop()) ;
                    }
                    if (pendingOps.peek() == '+') {
                        postfix.push('+') ;
                    } else {
                        pendingOps.push('+') ;
                    }
                }
            } else /* Object is an Integer */ {
                postfix.push(token);
            }
        }
        while(pendingOps.peek() != '(') {
            postfix.push(pendingOps.pop()) ;
        }
        return postfix ;
    }
    
    // Transform an postfix list, such as [ 13 17 + 19 * 21 + ]
    // into a result
    private static int doPostfixEval(Deque<Object> postfix) {
        // evaluation stack
        Deque<Integer> evalStack = new LinkedList<>() ;
        Iterator<Object> revStack = postfix.descendingIterator() ;
        while (revStack.hasNext()) {
            Object token = revStack.next() ;
            if (token instanceof Character) {
                // Must be + or -
                int operand1 = evalStack.pop() ;
                int operand2 = evalStack.pop() ;
                int result = (Character)token == '+' 
                        ? operand1 + operand2
                        : operand1 * operand2 ;
                evalStack.push(result) ;
            } else /* token instanceof Integer */ {
                evalStack.push((Integer)token) ;
            }
        }
        return evalStack.pop() ;      
    }
    
    private static void debugPrintStack(Deque<Object> stack) {
        Iterator<Object> revStack = stack.descendingIterator() ;
        System.out.print("[ ") ;
        while (revStack.hasNext()) {
            Object e = revStack.next() ;
            if (e instanceof Integer) {
                System.out.print((int)e) ;
            } else if (e instanceof Character) {
                System.out.print((char)e) ;
            } else /* This better no happen */ {
                System.out.print('?') ;
            }
            System.out.print(' ');
        }
        System.out.println(']') ;        
    }
    
    final static boolean debugEval = true ;
    
    /**
     * @param expr The expression to evaluate
     * @return     Value of the evaluated expression
     * @throws     IllegalArgumentException When expression has improper syntax 
     */
    public static int evaluateExpression(String expr)
            throws IllegalArgumentException {
        try {
            Deque<Object> infixList = string2infix(expr) ;
            if (debugEval) {
                System.out.print("Infix: ") ;
                debugPrintStack(infixList) ;
            }
            syntaxCheck(infixList) ;
            Deque<Object> rpnList = infix2postfix(infixList) ;
            if (debugEval) {
                System.out.print("Postfix: ") ;
                debugPrintStack(rpnList) ;
            }
            return doPostfixEval(rpnList) ;
        } catch (IllegalArgumentException iae) {
            throw iae ;
        }
    }
    
    private final static String[] expressions = {
        "15 + e",
        "202",
        "+",
        "181 182" ,
        "5 + 7 )",
        "()",
        "5 + (",
        "5 +" ,
        "15*17+ 23" ,
        "15+17*23",
        "(15+17)*23",
        "((15 + 17))",
        "(13 + 17 * 19 + 23) + 29 * (31 * 37 + 41) + 43 * (((47))) + 53"
   } ;
   
    
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        for (String e : expressions) {
            System.out.println("Evaluating: " + e) ;
            try {
                System.out.println("Result: " + evaluateExpression(e)) ;
            } catch(IllegalArgumentException ae) {
                System.out.println("Bad expression: " + ae.getMessage()) ;
            }
            System.out.println() ;
        }
    }

}
